import numpy as np
import scipy.optimize

f_phi = scipy.stats.norm.cdf    #   Cumulative standard normal distribution function

def my_log(x):
    """
            To avoid the error message with 0 values for log function
    """
    if x < 1.0e-323:
        return -744.0
    else:
        return np.log(x)

class LLog:
    """
            Negative log likelihood
    """
    def __init__(self, X, Res):
        self.N = len(X)
        self.X = X
        self.Res = Res

    def v(self, z):
        mu = z[0]
        sgm = z[1] ** 2         #   Squared to avoid negative values
        C = z[2] ** 2           #   Squared to avoid negative values
        v = 0
        for t in range(self.N):
            if self.Res[t] == 4:
                v += -my_log(f_phi((self.X[t] - mu - C) / sgm))
            elif self.Res[t] == 3:
                v += -my_log(f_phi((self.X[t] - mu) / sgm) -
                             f_phi((self.X[t] - mu - C) / sgm))
            elif self.Res[t] == 2:
                v += -my_log(f_phi((self.X[t] - mu + C) / sgm) -
                             f_phi((self.X[t] - mu) / sgm))
            else:
                v += -my_log(1.0 - f_phi((self.X[t] - mu + C) / sgm))
        return v

def calcMAP(X, Res, PSEMed, SgmMed, CMed):
    """
            Calculate MAP estimates
    """
    llog = LLog(X, Res)         #   Negative log likelihood

    z = np.empty(3)
    z[0] = PSEMed               #   Initial value for mu
    z[1] = SgmMed ** 0.5        #   Initial value for square root of sigma
    z[2] = CMed ** 0.5          #   Initial value for square root of C

    res = scipy.optimize.minimize(llog.v, z, method = 'Nelder-Mead',
                        options = {'xatol': 1.0e-7, 'fatol': 1.0e-12})

    muMAP = res.x[0]
    sgmMAP = res.x[1] ** 2
    CMAP = res.x[2] ** 2

    return muMAP, sgmMAP, CMAP


class PF:
    """
            Psychometric function with the parameters mu and sigma
    """
    def __init__(self, mu, sgm):
        self.mu = mu
        self.sgm = sgm

    def v(self, x):
        return f_phi((x - self.mu) / self.sgm)

def bisect( f, c, Lb, Ub ):
    """
            Return the value root so that f(root) = c
    """
    s = 1.0 if f(Ub) > f(Lb) else -1.0      #   s = 1 when f(x) is an increasing functin
    while ((abs(Ub - Lb) > (1.0e-15) * (abs(Ub) + abs(Lb)))
        and
          (abs(Ub) + abs(Lb) > 1.0e-14)):
              m = 0.5 * (Ub + Lb)
              if (s * (f(m) - c)) > 0.0:
                  Ub = m
              else:
                  Lb = m
 
    return 0.5 * (Ub + Lb)
 
 

def my_kernel(x, y, w_k):
    """
            Kernel function with width w_k
    """
    return np.exp(-((x - y) / w_k) ** 2)


def k_ave(v, X, W, w_k):
    """
            Average by kernel function my_kernel
    """
    n = len(X)
    num = 0.0
    den = 0.0
    for i in range(n):
        k_v = my_kernel(v, X[i], w_k)
        num += W[i] * k_v
        den += k_v
    return num / den


